@@ -34,7 +34,8 @@
 #include <QtCore/QDataStream>
 #include <QtCore/QByteArray>
 #include <QtCore/QStringBuilder> // % operator for QString
-#include <QCoreApplication>
+#include <QElapsedTimer>
+#include <QGuiApplication>
 #include <QIcon>
 #include <QImage>
 #include <QMovie>
@@ -48,6 +49,10 @@
 #include <ksharedconfig.h>
 #include <QtDBus/QDBusConnection>
 #include <QDBusMessage>
+#include <QSvgRenderer>
+#include <QXmlStreamReader>
+#include <QXmlStreamWriter>
+#include <QCryptographicHash>
 
 // kdeui
 #include "kicontheme.h"
@@ -57,8 +62,38 @@
 //kwidgetsaddons
 #include <kpixmapsequence.h>
 
+#include <kcolorscheme.h>
+#include <kcompressiondevice.h>
+
+namespace {
 // Used to make cache keys for icons with no group. Result type is QString*
-Q_GLOBAL_STATIC_WITH_ARGS(QString, NULL_EFFECT_FINGERPRINT, (QString::fromLatin1("noeffect")))
+QString NULL_EFFECT_FINGERPRINT()
+{
+    return QStringLiteral("noeffect");
+}
+
+QString STYLESHEET_TEMPLATE()
+{
+    return QStringLiteral(".ColorScheme-Text {\
+color:%1;\
+}\
+.ColorScheme-Background{\
+color:%2;\
+}\
+.ColorScheme-Highlight{\
+color:%3;\
+}\
+.ColorScheme-PositiveText{\
+color:%4;\
+}\
+.ColorScheme-NeutralText{\
+color:%5;\
+}\
+.ColorScheme-NegativeText{\
+color:%6;\
+}");
+}
+}
 
 /**
  * Checks for relative paths quickly on UNIX-alikes, slowly on everything else.
@@ -80,6 +115,37 @@ struct PixmapWithPath {
     QString path;
 };
 
+/**
+ * Function to convert an uint32_t to AARRGGBB hex values.
+ *
+ * W A R N I N G !
+ * This function is for internal use!
+ */
+KICONTHEMES_EXPORT void uintToHex(uint32_t colorData, QChar *buffer)
+{
+    static const char hexLookup[] = "0123456789abcdef";
+    buffer += 7;
+    uchar *colorFields = reinterpret_cast<uchar*>(&colorData);
+
+    for (int i = 0; i < 4; i++) {
+        *buffer-- = hexLookup[*colorFields & 0xf];
+        *buffer-- = hexLookup[*colorFields >> 4];
+        colorFields++;
+    }
+}
+
+static QString paletteId(const QPalette &pal)
+{
+    // 8 per color. We want 3 colors thus 8*3=24.
+    QString buffer(24, Qt::Uninitialized);
+
+    uintToHex(pal.text().color().rgba(), buffer.data());
+    uintToHex(pal.highlight().color().rgba(), buffer.data() + 8);
+    uintToHex(pal.highlightedText().color().rgba(), buffer.data() + 16);
+
+    return buffer;
+}
+
 /*** KIconThemeNode: A node in the icon theme dependancy tree. ***/
 
 class KIconThemeNode
@@ -132,6 +198,9 @@ struct KIconGroup {
     int size;
 };
 
+extern KICONTHEMES_EXPORT int kiconloader_ms_between_checks;
+KICONTHEMES_EXPORT int kiconloader_ms_between_checks = 5000;
+
 /*** d pointer for KIconLoader. ***/
 class KIconLoaderPrivate
 {
@@ -252,13 +321,22 @@ public:
     QString makeCacheKey(const QString &name, KIconLoader::Group group, const QStringList &overlays,
                          int size, int state) const;
 
+    /**
+     * @internal
+     * If the icon is an SVG file, process it generating a stylesheet
+     * following the current color scheme. in this case the icon can use named colors
+     * as text color, background color, highlight color, positive/neutral/negative color
+     * @see KColorScheme
+     */
+    QByteArray processSvg(const QString &path, KIconLoader::States state) const;
+
     /**
      * @internal
      * Creates the QImage for @p path, using SVG rendering as appropriate.
      * @p size is only used for scalable images, but if non-zero non-scalable
      * images will be resized anyways.
      */
-    QImage createIconImage(const QString &path, int size = 0);
+    QImage createIconImage(const QString &path, int size = 0, KIconLoader::States state = KIconLoader::DefaultState);
 
     /**
      * @internal
@@ -284,6 +362,15 @@ public:
      */
     void _k_refreshIcons(int group);
 
+    bool shouldCheckForUnknownIcons()
+    {
+        if (mLastUnknownIconCheck.isValid() && mLastUnknownIconCheck.elapsed() < kiconloader_ms_between_checks) {
+            return false;
+        }
+        mLastUnknownIconCheck.start();
+        return true;
+    }
+
     KIconLoader *const q;
 
     QStringList mThemesInTree;
@@ -307,7 +394,8 @@ public:
 
     void drawOverlays(const KIconLoader *loader, KIconLoader::Group group, int state, QPixmap &pix, const QStringList &overlays);
 
-    QSet<QString> mAvailableIcons;
+    QHash<QString, bool> mIconAvailability; // icon name -> true (known to be available) or false (known to be unavailable)
+    QElapsedTimer mLastUnknownIconCheck; // recheck for unknown icons after kiconloader_ms_between_checks
 };
 
 class KIconLoaderGlobalData : public QObject
@@ -455,7 +543,7 @@ void KIconLoaderPrivate::_k_refreshIcons(int group)
 {
     KSharedConfig::openConfig()->reparseConfiguration();
     q->newIconLoader();
-    mAvailableIcons.clear();
+    mIconAvailability.clear();
     emit q->iconChanged(group);
 }
 
@@ -611,7 +699,6 @@ void KIconLoaderPrivate::addBaseThemes(KIconThemeNode *node, const QString &appn
     // present in an inherited theme, and hicolor goes last.
 
     addInheritedThemes(node, appname);
-    addThemeByName(KIconTheme::defaultThemeName(), appname);
     addThemeByName(QStringLiteral("hicolor"), appname);
 }
 
@@ -666,8 +753,8 @@ void KIconLoaderPrivate::addExtraDesktopThemes()
         const QStringList lst = dir.entryList(QStringList(QStringLiteral("default.*")), QDir::Dirs);
         QStringList::ConstIterator it2;
         for (it2 = lst.begin(); it2 != lst.end(); ++it2) {
-            if (!QFile::exists(*it + *it2 + "/index.desktop")
-                    && !QFile::exists(*it + *it2 + "/index.theme")) {
+            if (!QFileInfo::exists(*it + *it2 + "/index.desktop")
+                    && !QFileInfo::exists(*it + *it2 + "/index.theme")) {
                 continue;
             }
             //TODO: Is any special handling required for NTFS symlinks?
@@ -764,22 +851,84 @@ QString KIconLoaderPrivate::makeCacheKey(const QString &name, KIconLoader::Group
            % QLatin1Char('_')
            % overlays.join(QStringLiteral("_"))
            % (group >= 0 ? mpEffect.fingerprint(group, state)
-              : *NULL_EFFECT_FINGERPRINT());
+              : NULL_EFFECT_FINGERPRINT())
+           % QLatin1Char('_')
+           % paletteId(qApp->palette())
+           % (q->theme() && q->theme()->followsColorScheme() && state == KIconLoader::SelectedState ? QStringLiteral("_selected") : QString());
 }
 
-QImage KIconLoaderPrivate::createIconImage(const QString &path, int size)
+QByteArray KIconLoaderPrivate::processSvg(const QString &path, KIconLoader::States state) const
 {
-    QImageReader reader(path);
+    QScopedPointer<QIODevice> device;
 
-    if (!reader.canRead()) {
+    if (path.endsWith(QLatin1String("svgz"))) {
+        device.reset(new KCompressionDevice(path, KCompressionDevice::GZip));
+    } else {
+        device.reset(new QFile(path));
+    }
+
+    if (!device->open(QIODevice::ReadOnly)) {
+        return QByteArray();
+    }
+
+    const QPalette pal = qApp->palette();
+    KColorScheme scheme(QPalette::Active, KColorScheme::Window);
+    QString styleSheet = STYLESHEET_TEMPLATE().arg(
+        state == KIconLoader::SelectedState ? pal.highlightedText().color().name() : pal.windowText().color().name(),
+        state == KIconLoader::SelectedState ? pal.highlight().color().name() : pal.window().color().name(),
+        state == KIconLoader::SelectedState ? pal.highlightedText().color().name() : pal.highlight().color().name(),
+        scheme.foreground(KColorScheme::PositiveText).color().name(),
+        scheme.foreground(KColorScheme::NeutralText).color().name(),
+        scheme.foreground(KColorScheme::NegativeText).color().name());
+
+    QByteArray processedContents;
+    QXmlStreamReader reader(device.data());
+
+    QBuffer buffer(&processedContents);
+    buffer.open(QIODevice::WriteOnly);
+    QXmlStreamWriter writer(&buffer);
+    while (!reader.atEnd()) {
+        if (reader.readNext() == QXmlStreamReader::StartElement &&
+            reader.qualifiedName() == QLatin1String("style") &&
+            reader.attributes().value(QLatin1String("id")) == QLatin1String("current-color-scheme")) {
+            writer.writeStartElement(QLatin1String("style"));
+            writer.writeAttributes(reader.attributes());
+            writer.writeCharacters(styleSheet);
+            writer.writeEndElement();
+            while (reader.tokenType() != QXmlStreamReader::EndElement) {
+                reader.readNext();
+            }
+        } else if (reader.tokenType() != QXmlStreamReader::Invalid) {
+            writer.writeCurrentToken(reader);
+        }
+    }
+    buffer.close();
+
+    return processedContents;
+}
+
+QImage KIconLoaderPrivate::createIconImage(const QString &path, int size, KIconLoader::States state)
+{
+    //TODO: metadata in the theme to make it do this only if explicitly supported?
+    QScopedPointer<QImageReader> reader;
+    QBuffer buffer;
+
+    if (q->theme()->followsColorScheme() && (path.endsWith(QLatin1String("svg")) || path.endsWith(QLatin1String("svgz")))) {
+        buffer.setData(processSvg(path, state));
+        reader.reset(new QImageReader(&buffer));
+    } else {
+        reader.reset(new QImageReader(path));
+    }
+
+    if (!reader->canRead()) {
         return QImage();
     }
 
     if (size != 0) {
-        reader.setScaledSize(QSize(size, size));
+        reader->setScaledSize(QSize(size, size));
     }
 
-    return reader.read();
+    return reader->read();
 }
 
 void KIconLoaderPrivate::insertCachedPixmapWithPath(
@@ -886,6 +1035,21 @@ QString KIconLoaderPrivate::findMatchingIcon(const QString &name, int size) cons
 {
     const_cast<KIconLoaderPrivate *>(this)->initIconThemes();
 
+    // Do two passes through themeNodes.
+    //
+    // The first pass looks for an exact match in each themeNode one after the other.
+    // If one is found and it is an app icon then return that icon.
+    //
+    // In the next pass (assuming the first pass failed), it looks for
+    // generic fallbacks in each themeNode one after the other.
+
+    // In theory we should only do this for mimetype icons, not for app icons,
+    // but that would require different APIs. The long term solution is under
+    // development for Qt >= 5.8, QFileIconProvider calling QPlatformTheme::fileIcon,
+    // using QMimeType::genericIconName() to get the proper -x-generic fallback.
+    // Once everone uses that to look up mimetype icons, we can kill the fallback code
+    // from this method.
+
     foreach (KIconThemeNode *themeNode, links) {
         const QString path = themeNode->theme->iconPathByName(name, size, KIconLoader::MatchBest);
         if (!path.isEmpty()) {
@@ -893,18 +1057,15 @@ QString KIconLoaderPrivate::findMatchingIcon(const QString &name, int size) cons
         }
     }
 
-    bool genericFallback = name.endsWith(QLatin1String("-x-generic"));
+    if (name.endsWith(QLatin1String("-x-generic"))) {
+        return QString(); // no further fallback
+    }
+    bool genericFallback = false;
     QString path;
     foreach (KIconThemeNode *themeNode, links) {
         QString currentName = name;
 
         while (!currentName.isEmpty()) {
-            //qCDebug(KICONTHEMES) << "Looking up" << currentName;
-
-            path = themeNode->theme->iconPathByName(currentName, size, KIconLoader::MatchBest);
-            if (!path.isEmpty())
-                return path;
-
             if (genericFallback) {
                 // we already tested the base name
                 break;
@@ -934,6 +1095,16 @@ QString KIconLoaderPrivate::findMatchingIcon(const QString &name, int size) cons
                     break;
                 }
             }
+
+            if (currentName.isEmpty()) {
+                break;
+            }
+
+            //qCDebug(KICONTHEMES) << "Looking up" << currentName;
+            path = themeNode->theme->iconPathByName(currentName, size, KIconLoader::MatchBest);
+            if (!path.isEmpty()) {
+                return path;
+            }
         }
     }
     return path;
@@ -954,7 +1125,7 @@ QString KIconLoaderPrivate::locate(const QString &fileName)
     Q_FOREACH (const QString &dir, searchPaths) {
         const QString path = dir + '/' + fileName;
         if (QDir(dir).isAbsolute()) {
-            if (QFile::exists(path)) {
+            if (QFileInfo::exists(path)) {
                 return path;
             }
         } else {
@@ -1103,14 +1274,20 @@ QPixmap KIconLoader::loadIcon(const QString &_name, KIconLoader::Group group, in
     bool iconWasUnknown = false;
     QString path;
 
-    // icon.path would be empty for "unknown" icons, which should be searched for
-    // anew each time.
-    if (d->findCachedPixmapWithPath(key, pix, path) && !path.isEmpty()) {
+    if (d->findCachedPixmapWithPath(key, pix, path)) {
         if (path_store) {
             *path_store = path;
         }
 
-        return pix;
+        if (!path.isEmpty()) {
+            return pix;
+        } else {
+            // path is empty for "unknown" icons, which should be searched for
+            // anew regularly
+            if (!d->shouldCheckForUnknownIcons()) {
+                return canReturnNull ? QPixmap() : pix;
+            }
+        }
     }
 
     // Image is not cached... go find it and apply effects.
@@ -1145,7 +1322,7 @@ QPixmap KIconLoader::loadIcon(const QString &_name, KIconLoader::Group group, in
 
     QImage img;
     if (!path.isEmpty()) {
-        img = d->createIconImage(path, size);
+        img = d->createIconImage(path, size, (KIconLoader::States)state);
     }
 
     if (group >= 0) {
@@ -1578,10 +1755,17 @@ QPixmap KIconLoader::unknown()
 
 bool KIconLoader::hasIcon(const QString &name) const
 {
-    bool found = d->mAvailableIcons.contains(name);
-    if (!found && !iconPath(name, KIconLoader::Desktop, KIconLoader::MatchBest).isEmpty()) {
-        found = true;
-        d->mAvailableIcons.insert(name);
+    auto it = d->mIconAvailability.constFind(name);
+    const auto end = d->mIconAvailability.constEnd();
+    if (it != end && !it.value() && !d->shouldCheckForUnknownIcons()) {
+        return false; // known to be unavailable
+    }
+    bool found = it != end && it.value();
+    if (!found) {
+        if (!iconPath(name, KIconLoader::Desktop, KIconLoader::MatchBest).isEmpty()) {
+            found = true;
+        }
+        d->mIconAvailability.insert(name, found); // remember whether the icon is available or not
     }
     return found;
 }
